// 
// methods here from Sedgewick 7.1, 7.2. (Ascending order) 

import java.util.Vector;

public class QuickSort {
    private static long comparisons = 0;
    private static long exchanges   = 0;
 
    public static void quicksort(double[] a) {
        shuffle(a);                        // to guard against worst-case
        quicksort(a, 0, a.length - 1);
    }

    public static void quicksort(double[] a, String[] s) {
        shuffle(a, s);                        // to guard against worst-case
        quicksort(a, s, 0, a.length - 1);
    }

    
    public static void quicksort(double[] a, int left, int right) {
        if (right <= left) return;
        int i = partition(a, left, right);
        quicksort(a, left, i-1);
        quicksort(a, i+1, right);
    }

    public static void quicksort(double[] a, String[] s, int left, int right) {
        if (right <= left) return;
        int i = partition(a, s, left, right);
        quicksort(a, s ,left, i-1);
        quicksort(a, s ,i+1, right);
    }

    private static int partition(double[] a, int left, int right) {
        int i = left - 1;
        int j = right;
        while (true) {
            while (less(a[++i], a[right]))      // find item on left to swap
                ;                               // a[right] acts as sentinel
            while (less(a[right], a[--j]))      // find item on right to swap
                if (j == left) break;           // don't go out-of-bounds
            if (i >= j) break;                  // check if pointers cross
            exch(a, i, j);                      // swap two elements into place
        }
        exch(a, i, right);                      // swap with partition element
        return i;
    }
    
    private static int partition(double[] a, String[] s, int left, int right) {
        int i = left - 1;
        int j = right;
        while (true) {
            while (less(a[++i], a[right]))      // find item on left to swap
                ;                               // a[right] acts as sentinel
            while (less(a[right], a[--j]))      // find item on right to swap
                if (j == left) break;           // don't go out-of-bounds
            if (i >= j) break;                  // check if pointers cross
            exch(a, s, i, j);                      // swap two elements into place
        }
        exch(a, s, i, right);                      // swap with partition element
        return i;
    }
    
    // is x < y ?
    private static boolean less(double x, double y) {
        comparisons++;
        return (x < y);
    }

    // exchange a[i] and a[j]
    private static void exch(double[] a, int i, int j) {
        exchanges++;
        double swap = a[i];
        a[i] = a[j];
        a[j] = swap;
    }

    
    // shuffle the array a
    private static void shuffle(double[] a) {
        int N = a.length;
        for (int i = 0; i < N; i++) {
            int r = i + (int) (Math.random() * (N-i));   // between i and N-1
            exch(a, i, r);
        }
    }
    
        // exchange a[i] and a[j]
    private static void exch(double[] a, String []s, int i, int j) {
        exchanges++;
        double swap = a[i];
        a[i] = a[j];
        a[j] = swap;

	String ssw = s[i];
	s[i] = s[j];
	s[j] = ssw;
    }

    
    // shuffle the array a
    private static void shuffle(double[] a, String[] s) {
        int N = a.length;
        for (int i = 0; i < N; i++) {
            int r = i + (int) (Math.random() * (N-i));   // between i and N-1
            exch(a, s, i, r);
        }
    }
}


